/* protocols.c
	Suuport functions for sendip.c
*/

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/udp.h>
#include <netinet/tcp.h>
#include <stdlib.h>
#include "csum.h"

/* glibc renamed everything in the tcp_hdr structure so we use macros to
	access the elements.  This is sick. */
#ifdef __GLIBC__
#define SPORT  source
#define DPORT  dest
#define SEQ    seq
#define ACK    ack_seq
#define RES    res1
#define OFFSET doff
#define WINDOW window
#define CHECK  check
#define URG    urg_ptr
#define SETRES(tcp,i) tcp->res2=(i);
#define SETURG(tcp,i) tcp->urg=(i);
#define SETACK(tcp,i) tcp->ack=(i);
#define SETPSH(tcp,i) tcp->psh=(i);
#define SETRST(tcp,i) tcp->rst=(i);
#define SETFIN(tcp,i) tcp->fin=(i);
#define SETSYN(tcp,i) tcp->syn=(i);
#else     /* Now the libc5 versions */
#define SPORT  th_sport
#define DPORT  th_dport
#define SEQ    th_seq
#define ACK    th_ack
#define RES    th_x2
#define OFFSET th_off
#define WINDOW th_win
#define CHECK  th_sum
#define URG    th_urp
#define SETRES(tcp,i) tcp->th_flags &= (i)<<6;
#define SETURG(tcp,i) (i)?(tcp->th_flags|=TH_URG):(tcp->th_flags&=~TH_URG);
#define SETACK(tcp,i) (i)?(tcp->th_flags|=TH_ACK):(tcp->th_flags&=~TH_ACK);
#define SETPSH(tcp,i) (i)?(tcp->th_flags|=TH_PUSH):(tcp->th_flags&=~TH_PUSH);
#define SETRST(tcp,i) (i)?(tcp->th_flags|=TH_RST):(tcp->th_flags&=~TH_RST);
#define SETFIN(tcp,i) (i)?(tcp->th_flags|=TH_FIN):(tcp->th_flags&=~TH_FIN);
#define SETSYN(tcp,i) (i)?(tcp->th_flags|=TH_SYN):(tcp->th_flags&=~TH_SYN);
#endif

extern void usage_exit(const char *err);

void ip_initialize(struct iphdr *iph);
void icmp_initialize(struct icmphdr *icp);
void tcp_initialize(struct tcphdr *tcp);
void udp_initialize(struct udphdr *udp);

void do_ipopt(struct iphdr *iph, char *opt, char *arg);
void do_icmpopt(struct icmphdr *icp, char *opt, char *arg);
void do_tcpopt(struct tcphdr *tcp, char *opt, char *arg);
void do_udpopt(struct udphdr *udp, char *opt, char *arg);

void ip_finalize(struct iphdr *iph, int length, struct sockaddr_in *to);
void icmp_finalize(struct iphdr *iph, struct icmphdr *icp, int length);
void tcp_finalize(struct iphdr *iph, struct tcphdr *tcp, int length);
void udp_finalize(struct iphdr *iph, struct udphdr *udp, int length);

void ip_initialize(struct iphdr *iph) {
	iph->version=  (unsigned int)0x4;
	iph->ihl=      (unsigned int)0x5;
	iph->tos=          (u_int8_t)0x00;
	iph->tot_len=     (u_int16_t)0xFFFF;        /* automatic */
	iph->id=          (u_int16_t)rand();
	iph->frag_off=    (u_int16_t)0x0000;
	iph->ttl=          (u_int8_t)0xFF;
	iph->protocol=     (u_int8_t)0xFF;          /* automatic */
	iph->check=       (u_int16_t)0xFFFF;        /* automatic */
	iph->saddr=                 INADDR_NONE;    /* automatic */
	iph->daddr=                 INADDR_NONE;    /* automatic */
}

void icmp_initialize(struct icmphdr *icp) {
	icp->type=                  ICMP_ECHO;
	icp->code=         (u_int8_t)0x00;
	icp->checksum=    (u_int16_t)0xFFFF;        /* automatic */
}

void tcp_initialize(struct tcphdr *tcp) {
	tcp->SPORT =      (u_int16_t)0x0000;
	tcp->DPORT =      (u_int16_t)0x0000;
	tcp->SEQ =        (u_int32_t)rand();
	tcp->ACK =        (u_int32_t)0x00000000;
	tcp->OFFSET =     (u_int16_t)0x5;
	tcp->RES =        (u_int16_t)0x0;
#ifdef __GLIBC__
	tcp->res2 =       (u_int16_t)0x0;
	tcp->urg=tcp->ack=tcp->psh=tcp->rst=tcp->fin=(u_int16_t)0;
	tcp->syn=         (u_int16_t)1;
#else
	tcp->th_flags =   (u_int16_t)TH_SYN;
#endif
	tcp->WINDOW =     (u_int16_t)65535;
	tcp->CHECK =      (u_int16_t)0xFFFF;  /* automatic */
	tcp->URG =        (u_int16_t)0x0000;
}

void udp_initialize(struct udphdr *udp) {
	udp->source =     (u_int16_t)0x0000;
	udp->dest =       (u_int16_t)0x0000;
	udp->len =        (u_int16_t)0xFFFF;        /* automatic */
	udp->check =      (u_int16_t)0xFFFF;        /* automatic */
}

void do_ipopt(struct iphdr *iph, char *opt, char *arg) {
	switch(opt[2]) {
	case 'h':
		iph->ihl = (unsigned int)strtoul(arg, (char **)NULL, 0) & 0xF;
		break;
	case 'y':
		iph->tos = (u_int8_t)strtoul(arg, (char **)NULL, 0);
		break;
	case 'l':
		iph->tot_len = htons((u_int16_t)strtoul(arg, (char **)NULL, 0));
		break;
	case 'i':
		iph->id = htons((u_int16_t)strtoul(arg, (char **)NULL, 0));
		break;
	case 't':
		iph->ttl = (u_int8_t)strtoul(arg, (char **)NULL, 0);
		break;
	case 'p':
		iph->protocol = (u_int8_t)strtoul(arg, (char **)NULL, 0);
		break;
	case 'c':
		iph->check = htons((u_int16_t)strtoul(arg, (char **)NULL, 0));
		break;
	case 's':
		iph->saddr = inet_addr(arg);
		break;
	case 'd':
		iph->daddr = inet_addr(arg);
		break;
	case 'f':
		switch(opt[3]) {
		case 'r':
			if(*arg != '0' && *arg != '1') {
				usage_exit("IP reserved flag must be 0 or 1\n");
			}
			*arg-='0';
			if(*arg) {
				iph->frag_off |= (u_int16_t)0x8000;
			} else {
				iph->frag_off &= (u_int16_t)0x7FFF;
			}
			break;
		case 'd':
			if(*arg != '0' && *arg != '1') {
				usage_exit("IP don't fragment flag must be 0 or 1\n");
			}
			*arg-='0';
			if(*arg) {
				iph->frag_off |= (u_int16_t)0x4000;
			} else {
				iph->frag_off &= (u_int16_t)0xBFFF;
			}
			break;
		case 'm':
			if(*arg != '0' && *arg != '1') {
				usage_exit("IP more fragments flag must be 0 or 1\n");
			}
			*arg-='0';
			if(*arg) {
				iph->frag_off |= (u_int16_t)0x2000;
			} else {
				iph->frag_off &= (u_int16_t)0xDFFF;
			}
			break;
		case '\0':
			iph->frag_off &= (u_int16_t)0xE000;
			iph->frag_off |= (u_int16_t)strtoul(arg, (char **)NULL, 0) & 
				(u_int16_t)0x1FFF;
			break;
		default:
			usage_exit("unknown IP flag\n");
			break;
		}
		break;
	default:
		usage_exit("unknown IP option\n");
		break;
	}
}

void do_icmpopt(struct icmphdr *icp, char *opt, char *arg) {
	switch(opt[2]) {
	case 't':
		icp->type = (u_int8_t)strtoul(arg, (char **)NULL, 0);
		break;
	case 'd':
		icp->code = (u_int8_t)strtoul(arg, (char **)NULL, 0);
		break;
	case 'c':
		icp->checksum = htons((u_int16_t)strtoul(arg, (char **)NULL, 0));
		break;
	default:
		usage_exit("unknown ICMP option\n");
		break;
	}
}

void do_tcpopt(struct tcphdr *tcp, char *opt, char *arg) {
	switch(opt[2]) {
	case 's':
		tcp->SPORT = htons((u_int16_t)strtoul(arg, (char **)NULL, 0));
		break;
	case 'd':
		tcp->DPORT = htons((u_int16_t)strtoul(arg, (char **)NULL, 0));
		break;
	case 'n':
		tcp->SEQ = htonl((u_int32_t)strtoul(arg, (char **)NULL, 0));
		break;
	case 'a':
		tcp->ACK = htonl((u_int32_t)strtoul(arg, (char **)NULL, 0));
		SETACK(tcp,1);
		break;
	case 't':
		tcp->OFFSET = (u_int16_t)strtoul(arg, (char **)NULL, 0) &0xF;
		break;
	case 'r':
		tcp->RES = (u_int16_t)(strtoul(arg, (char **)NULL, 0) & 0x003C)>>2;
		SETRES(tcp,(u_int16_t)strtoul(arg, (char **)NULL, 0) & 0x0003);
		break;
	case 'f':
		if(*arg != '0' && *arg != '1') {
			usage_exit("TCP flag must be 0 or 1\n");
		}
		*arg-=0x30;
		switch(opt[3]) {
		case 'u':
			SETURG(tcp,(u_int16_t)*arg&1);
			break;
		case 'a':
			SETACK(tcp,(u_int16_t)*arg&1);
			break;
		case 'p':
			SETPSH(tcp,(u_int16_t)*arg&1);
			break;
		case 'r':
			SETRST(tcp,(u_int16_t)*arg&1);
			break;
		case 's':
			SETSYN(tcp,(u_int16_t)*arg&1);
			break;
		case 'f':
			SETFIN(tcp,(u_int16_t)*arg&1);
			break;
		default:
			usage_exit("TCP flage not known\n");
			break;
		}
		break;
	case 'w':
		tcp->WINDOW = htons((u_int16_t)strtoul(arg, (char **)NULL, 0));
		break;
	case 'c':
		tcp->CHECK = htons((u_int16_t)strtoul(arg, (char **)NULL, 0));
		break;
	case 'u':
		tcp->URG = htons((u_int16_t)strtoul(arg, (char **)NULL, 0));
		SETURG(tcp,1);
		break;
	default:
		usage_exit("unknown TCP options\n");
		break;
	}
}

void do_udpopt(struct udphdr *udp, char *opt, char *arg) {
	switch(opt[2]) {
	case 's':
		udp->source = htons((u_int16_t)strtoul(arg, (char **)NULL, 0));
		break;
	case 'd':
		udp->dest = htons((u_int16_t)strtoul(arg, (char **)NULL, 0));
		break;
	case 'l':
		udp->len = htons((u_int16_t)strtoul(arg, (char **)NULL, 0));
		break;
	case 'c':
		udp->check = htons((u_int16_t)strtoul(arg, (char **)NULL, 0));
		break;
	default:
		usage_exit("unknown UDP options\n");
		break;
	}
}

void ip_finalize(struct iphdr *iph, int length, struct sockaddr_in *to) {
	if(iph->tot_len==(u_int16_t)0xFFFF) iph->tot_len = htons(length);
	if(iph->protocol == (u_int8_t)0xFF) iph->protocol=0;
	if(iph->check == (u_int16_t)0xFFFF) {
		iph->check = 0;
		iph->check = csum((unsigned short *)iph, sizeof(struct iphdr));
	}
	if(iph->saddr == INADDR_NONE) iph->saddr = inet_addr("127.0.0.1");
	if(iph->daddr == INADDR_NONE) iph->daddr = to->sin_addr.s_addr;
}

void icmp_finalize(struct iphdr *iph, struct icmphdr *icp, int length) {
	if(iph->protocol == (u_int8_t)0xFF) iph->protocol = IPPROTO_ICMP;
	if(icp->checksum == (u_int16_t)0xFFFF) {
		icp->checksum = 0;
		icp->checksum = csum((unsigned short *)icp,
									length-sizeof(struct iphdr));
	}
}

void tcp_finalize(struct iphdr *iph, struct tcphdr *tcp, int length) {
	if(iph->protocol == (u_int8_t)0xFF) iph->protocol = IPPROTO_TCP;
	if(tcp->CHECK == (u_int16_t)0xFFFF) {
		tcp->CHECK = 0;
		tcp->CHECK = tcpcsum((unsigned char *)iph, (unsigned short *)tcp,
								length-sizeof(struct iphdr));
	}
}

void udp_finalize(struct iphdr *iph, struct udphdr *udp, int length) {
	if(udp->len==(u_int16_t)0xFFFF)
		udp->len = htons(length-sizeof(struct iphdr));
	if(iph->protocol == (u_int8_t)0xFF) iph->protocol = IPPROTO_UDP;
	if(udp->check == (u_int16_t)0xFFFF) {
		udp->check = 0;
		udp->check = udpcsum((unsigned char *)iph, (unsigned short *)udp,
									length-sizeof(struct iphdr));
	}
}
